@@ -1,4 +1,4 @@
-/*  $Id: BoolExpr.cpp,v 1.20 2006/02/05 03:26:11 sarrazip Exp $
+/*  $Id: BoolExpr.cpp,v 1.22 2008/10/09 03:14:05 sarrazip Exp $
     BoolExpr.cpp - Boolean expression binary tree node
 
     boolstuff - Disjunctive Normal Form boolean expression library
@@ -27,6 +27,12 @@
 #include <set>
 
 
+// Define this to enable traces that help in leak detection
+// when used with script leak-detector.pl.
+//
+//#define BOOLSTUFF_LEAK_DETECTION 1
+
+
 template <class T>
 BoolExpr<T>::BoolExpr(const T &initValue /*= T()*/)
   : type(VALUE),
@@ -34,6 +40,9 @@ BoolExpr<T>::BoolExpr(const T &initValue /*= T()*/)
     left(NULL),
     right(NULL)
 {
+    #ifdef BOOLSTUFF_LEAK_DETECTION
+    std::cerr << "CONS " << (void *) this << " " << type << " " << value << std::endl;
+    #endif
 }
 
 
@@ -44,6 +53,9 @@ BoolExpr<T>::BoolExpr(Type t, BoolExpr *l, BoolExpr *r)
     left(l),
     right(r)
 {
+    #ifdef BOOLSTUFF_LEAK_DETECTION
+    std::cerr << "CONS " << (void *) this << " " << type << " " << value << std::endl;
+    #endif
     assert(type == AND || type == OR || type == NOT);
 }
 
@@ -51,6 +63,9 @@ BoolExpr<T>::BoolExpr(Type t, BoolExpr *l, BoolExpr *r)
 template <class T>
 BoolExpr<T>::~BoolExpr()
 {
+    #ifdef BOOLSTUFF_LEAK_DETECTION
+    std::cerr << "DEST " << (void *) this << " " << type << " " << value << std::endl;
+    #endif
     delete left;
     delete right;
 }
@@ -212,7 +227,7 @@ BoolExpr<T>::getDNFTermRoots(OutputIter dest) const
 	    assert(right != NULL);
 	    assert((left == NULL) == (type == NOT));
 	case VALUE:
-	    *dest++ = this;
+	    *dest++ = const_cast<BoolExpr<T> *>(this);
 	    return dest;
 
 	default:
@@ -364,30 +379,87 @@ BoolExpr<T>::getDisjunctiveNormalForm(BoolExpr<T> *root)
     if (dnfRoot == NULL)
 	return dnfRoot;
 
-    std::vector<const BoolExpr<T> *> termRoots;
+    // Get the terms of the DNF:
+    std::vector<BoolExpr<T> *> termRoots;
     dnfRoot->getDNFTermRoots(std::inserter(termRoots, termRoots.end()));
 
+    // If only one term, consider it "useful",
+    // because we must return something:
+    if (termRoots.size() <= 1)
+    {
+	// Simplify terms of the form x & x.
+	simplifyXAndXTerms(termRoots);
+	return termRoots.front();
+    }
+
     // Determine which terms are useful:
     std::vector<BoolExpr<T> *> usefulTerms;
-    typename std::vector<const BoolExpr<T> *>::const_iterator it;
+    std::vector<bool> isUsefulTerm;
+    typename std::vector<BoolExpr<T> *>::const_iterator it;
     for (it = termRoots.begin(); it != termRoots.end(); it++)
     {
 	assert(*it != NULL);
-	if ((*it)->isDNFTermUseful())
+	bool u = (*it)->isDNFTermUseful();
+	isUsefulTerm.push_back(u);
+	if (u)
 	    usefulTerms.push_back(const_cast<BoolExpr<T> *>(*it));
     }
+    assert(isUsefulTerm.size() == termRoots.size());
 
     // Nothing to do if all the terms are useful:
     if (usefulTerms.size() == termRoots.size())
 	return dnfRoot;
 
     destroyDNFOrNodes(dnfRoot);
-    // dnfRoot now invalid
+    // dnfRoot now invalid; termRoots holds the term trees
 
+    // If all the terms are useless (e.g., a&!a | b&!b),
+    // return the first one, because we must return something.
+    if (usefulTerms.size() == 0)
+    {
+	// Destroy all term trees except the first one.
+	for (it = termRoots.begin(), it++; it != termRoots.end(); it++)
+	    delete *it;
+
+	// Return the first term.
+	return termRoots.front();
+    }
+
+    // Destroy the trees in 'termRoots' that are not useful:
+    for (it = termRoots.begin(); it != termRoots.end(); it++)
+	if (!isUsefulTerm[it - termRoots.begin()])
+	    delete *it;
+
+    // Simplify terms of the form x & x.
+    simplifyXAndXTerms(usefulTerms);
+
+    // Join the useful terms with OR nodes:
     return joinTreesWithOrNodes(usefulTerms);
 }
 
 
+template <class T>
+/*static*/
+void
+BoolExpr<T>::simplifyXAndXTerms(std::vector<BoolExpr<T> *> &terms)
+{
+    typename std::vector<BoolExpr<T> *>::iterator jt;
+    for (jt = terms.begin(); jt != terms.end(); jt++)
+    {
+	BoolExpr<T> *term = *jt;
+	if (term->type == AND
+		&& term->left->type == VALUE
+		&& term->right->type == VALUE
+		&& term->left->value == term->right->value)
+	{
+	    *jt = term->left;
+	    term->left = NULL;
+	    delete term;  // destroys AND, and VALUE at right
+	}
+    }
+}
+	    
+
 template <class T>
 /*static*/
 void
@@ -415,7 +487,7 @@ BoolExpr<T>::joinTreesWithOrNodes(const std::vector<BoolExpr<T> *> &trees)
     if (trees.size() == 0)
 	return NULL;
     if (trees.size() == 1)
-	return trees[0];
+	return trees.front();
 
     typename std::vector<BoolExpr<T> *>::const_iterator it = trees.begin();
 
@@ -735,6 +807,7 @@ operator << (std::ostream &out, const std::set<T> &s)
 }
 
 
+// A "literal" is a variable name, possibly accompanied by a NOT operator.
 template <class T>
 class Literal
 {
@@ -771,11 +844,13 @@ BoolExpr<T>::negateDNF(BoolExpr<T> *root)
     root->getDNFTermRoots(std::inserter(termRoots, termRoots.end()));
     assert(termRoots.size() > 0);
 
+    // Re-express the DNF as a list of lists of literals.
+
     std::vector< std::vector< Literal<T> > > terms;
 
     typename std::vector<const BoolExpr<T> *>::const_iterator it;
-    size_t product = 1;
-    const size_t max = 10000;
+    size_t product = 1;  // will be the product of the sizes of the terms
+    const size_t max = 10000;  // maximum allowed product (to avoid using excessive memory)
     for (it = termRoots.begin(); it != termRoots.end(); it++)
     {
 	std::set<T> pos, neg;
@@ -803,14 +878,19 @@ BoolExpr<T>::negateDNF(BoolExpr<T> *root)
 	}
     }
 
+
+    // 'root' and 'termRoots' are not useful anymore.
+    // Only 'terms' is used from now on.
+    termRoots.clear();
+    delete root;
+    root = NULL;
+
+
     std::vector<size_t> indexVec(terms.size(), 0);
     size_t numTermsCreated = 0;
     BoolExpr<T> *disjunction = NULL;
-    for (bool first = true; ; )
+    for (;;)
     {
-	if (first)
-	    first = false;
-
 	BoolExpr<T> *conjunction = NULL;
 	std::set< Literal<T> > usedLiterals;
 	for (size_t i = 0; i < indexVec.size(); i++)
@@ -818,19 +898,22 @@ BoolExpr<T>::negateDNF(BoolExpr<T> *root)
 	    assert(indexVec[i] < terms[i].size());
 	    const Literal<T> &lit = terms[i][indexVec[i]];
 
-	    BoolExpr<T> *literal = new BoolExpr<T>(lit.value);
-	    if (lit.pos)
-		literal = new BoolExpr<T>(NOT, NULL, literal);
+	    // If 'lit' not already used in the conjunction:
 
-	    // Insert literal in conjunction if not already there:
-	    if (conjunction == NULL)
+	    if (usedLiterals.find(lit) == usedLiterals.end())
 	    {
-		conjunction = literal;
-		usedLiterals.insert(lit);
-	    }
-	    else if (usedLiterals.find(lit) == usedLiterals.end())
-	    {
-		conjunction = new BoolExpr<T>(AND, conjunction, literal);
+		// Create a node (or two) for 'lit' and add it
+		// to the current conjunction.
+
+		BoolExpr<T> *literal = new BoolExpr<T>(lit.value);
+		if (lit.pos)
+		    literal = new BoolExpr<T>(NOT, NULL, literal);
+
+		if (conjunction == NULL)
+		    conjunction = literal;
+		else
+		    conjunction = new BoolExpr<T>(AND, conjunction, literal);
+
 		usedLiterals.insert(lit);
 	    }
 	}
